/*
 * node_fragment.c
 *
 * Copyright 2022 Jonathan Schöbel <jonathan@Ubermos-2019>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "macro.h"
#include "error.h"
#include "log.h"

#include "data.h"
#include "text.h"
#include "validator.h"

#include "fragment.h"

#include "node_fragment.h"


static const struct fragment_methods methods = {
	(struct SH_Fragment * (*)(struct SH_Fragment *,
				  struct SH_Error *))
	SH_NodeFragment_deepcopy,

	(void (*)(struct SH_Fragment *, struct SH_Error *))
	SH_NodeFragment_free,

	(struct SH_Text * (*)(struct SH_Fragment *, enum HTML_MODE,
			      unsigned int, unsigned int, const char *,
			      struct SH_Error *))
	SH_NodeFragment_to_html
};


struct SH_Fragment *
SH_NodeFragment_new (const char * tag, struct SH_Data * data,
		     struct SH_Error * error)
{
	struct SH_NodeFragment * fragment;

	if (!SH_Validator_check_tag (data->validator, tag))
	{
		ERROR2 ("Tag %s is'nt valid.\n", tag);

		if (error != NULL)
		{
			error->type = VALUE_ERROR;
		}

		return NULL;
	}

	fragment = malloc (sizeof (struct SH_NodeFragment));
	if (fragment == NULL)
	{
		ERROR1 ("Memory allocation for "
			"SH_NodeFragment failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return NULL;
	}

	init_fragment (&(fragment->base), data, NODE, &methods);

	fragment->tag = strdup (tag);
	if (errno == ENOMEM)
	{
		ERROR1 ("Memory allocation for fragment tag failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		free (fragment);

		return NULL;
	}


	fragment->child_n = 0;
	fragment->childs = malloc (0);

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return (struct SH_Fragment *) fragment;
}

void
SH_NodeFragment_free (struct SH_NodeFragment * fragment,
		      struct SH_Error * error)
{
	size_t index;

	free (fragment->tag);

	for (index = 0; index < fragment->child_n; index++)
	{
		SH_Fragment_free (fragment->childs[index], error);
	}
	free (fragment->childs);

	free_fragment (&fragment->base);

	free (fragment);

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return;
}

struct SH_Fragment *
SH_NodeFragment_copy (struct SH_NodeFragment * fragment,
		      struct SH_Error * error)
{
	struct SH_NodeFragment * copy;

	copy = malloc (sizeof (struct SH_NodeFragment));
	if (copy == NULL)
	{
		ERROR1 ("Memory allocation for "
			"SH_NodeFragment failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return NULL;
	}

	copy_fragment (&(copy->base), &(fragment->base));

	copy->tag = strdup (fragment->tag);
	if (errno == ENOMEM)
	{
		ERROR1 ("Memory allocation for fragment tag failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		free (copy);

		return NULL;
	}


	copy->child_n = 0;
	copy->childs = malloc (0);

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return (struct SH_Fragment *) copy;
}

struct SH_Fragment *
SH_NodeFragment_deepcopy (struct SH_NodeFragment * fragment,
		          struct SH_Error * error)
{
	struct SH_NodeFragment * copy;
	struct SH_Fragment * child;
	size_t index;

	copy = malloc (sizeof (struct SH_NodeFragment));
	if (copy == NULL)
	{
		ERROR1 ("Memory allocation for "
			"SH_NodeFragment failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return NULL;
	}

	copy_fragment (&(copy->base), &(fragment->base));

	copy->tag = strdup (fragment->tag);
	if (errno == ENOMEM)
	{
		ERROR1 ("Memory allocation for fragment tag failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		free (copy);

		return NULL;
	}


	copy->child_n = fragment->child_n;
	copy->childs = malloc (sizeof (struct SH_NodeFragment *)
			       * fragment->child_n);

	if (copy->child_n != 0 && copy->childs == NULL)
	{
		ERROR1 ("Memory allocation for fragment child failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		free (copy->tag);
		free (copy);

		return NULL;
	}

	for (index = 0; index < fragment->child_n; index++)
	{
		child = SH_Fragment_copy (fragment->childs[index],
					  error);

		if (child == NULL)
		{
			copy->child_n = index;
			SH_NodeFragment_free (copy, NULL);

			return NULL;
		}

		copy->childs[index] = child;
	}

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return (struct SH_Fragment *) copy;
}

bool
SH_Fragment_is_NodeFragment (struct SH_Fragment * fragment)
{
	return get_type (fragment) == NODE;
}

char *
SH_NodeFragment_get_tag (struct SH_NodeFragment * fragment,
			 struct SH_Error * error)
{
	char * tag;

	tag = strdup (fragment->tag);

	if (errno == ENOMEM)
	{
		ERROR1 ("Memory allocation for fragment tag failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return NULL;
	}

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return tag;
}

struct SH_Fragment *
SH_NodeFragment_get_child (struct SH_NodeFragment * fragment, size_t index,
			   struct SH_Error * error)
{
	if (index >= fragment->child_n)
	{
		ERROR1 ("Fragment: Child index out of range.\n");

		if (error != NULL)
		{
			error->type = VALUE_ERROR;
		}

		return NULL;
	}

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return fragment->childs[index];
}

bool
SH_NodeFragment_is_child (struct SH_NodeFragment * fragment,
			  struct SH_Fragment * child)
{
	size_t index;

	for (index = 0; index < fragment->child_n; index++)
	{
		if (fragment->childs[index] == child)
		{
			return TRUE;
		}
	}
	return FALSE;
}

bool
SH_NodeFragment_is_descendant (struct SH_NodeFragment * fragment,
			       struct SH_Fragment * child)
{
	size_t index;

	for (index = 0; index < fragment->child_n; index++)
	{
		if (fragment->childs[index] == child
		|| (SH_Fragment_is_NodeFragment (child)
		    && SH_NodeFragment_is_descendant (
					(struct SH_NodeFragment *)
					fragment->childs[index],
					child)))
		{
			return TRUE;
		}
	}
	return FALSE;
}

bool
SH_NodeFragment_append_child (struct SH_NodeFragment * fragment,
			      struct SH_Fragment * child,
			      struct SH_Error * error)
{
	fragment->childs = realloc (fragment->childs,
				      sizeof (struct SH_Fragment *)
				      * (fragment->child_n + 1));

	if (errno == ENOMEM)
	{
		ERROR1 ("Memory allocation for fragment child failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return FALSE;
	}

	fragment->childs[fragment->child_n] = child;
	fragment->child_n++;

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return TRUE;
}

struct SH_Text *
SH_NodeFragment_to_html (struct SH_NodeFragment * fragment,
			 enum HTML_MODE mode,
			 unsigned int indent_base,
			 unsigned int indent_step,
			 char * indent_char,
			 struct SH_Error * error)
{
	struct SH_Text * html;
	struct SH_Text * child;
	struct SH_Text * indent_text;
	size_t index;

	html = SH_Text_new (error);
	if (error != NULL && error->type != SUCCESS)
	{
		return NULL;
	}

	if (mode == WRAP)
	{
		indent_text = SH_Text_new (error);
		if (indent_text == NULL)
		{
			SH_Text_free (html, NULL);
			return NULL;
		}

		for (index = 0; index < indent_base; index++)
		{
			if (!SH_Text_append_string (indent_text,
						    indent_char, error))
			{
				SH_Text_free (html, NULL);
				SH_Text_free (indent_text, NULL);
				return NULL;
			}
		}

		if (!SH_Text_append_text (html, indent_text, error))
		{
			SH_Text_free (html, NULL);
			SH_Text_free (indent_text, NULL);
			return NULL;
		}
	}

	if (!SH_Text_append_string (html, OPEN_TAG_BEGIN, error))
	{
		SH_Text_free (html, NULL);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text, NULL);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, fragment->tag, error))
	{
		SH_Text_free (html, NULL);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text, NULL);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, OPEN_TAG_END, error))
	{
		SH_Text_free (html, NULL);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text, NULL);
		}

		return NULL;
	}

	if (mode == WRAP)
	{
		if (!SH_Text_append_string (html, NEWLINE, error))
		{
			SH_Text_free (html, NULL);
			SH_Text_free (indent_text, NULL);
			return NULL;
		}
	}

	for (index = 0; index < fragment->child_n; index++)
	{
		child = SH_Fragment_to_html (fragment->childs[index],
					     mode,
					     indent_base + indent_step,
					     indent_step,
					     indent_char,
					     error);

		if (child == NULL)
		{
			SH_Text_free (html, NULL);

			if (mode == WRAP)
			{
				SH_Text_free (indent_text, NULL);
			}

			return NULL;
		}

		SH_Text_join (html, child);
	}

	if (mode == WRAP)
	{
		if (!SH_Text_append_text (html, indent_text, error))
		{
			SH_Text_free (html, NULL);
			SH_Text_free (indent_text, NULL);
			return NULL;
		}
	}

	if (!SH_Text_append_string (html, CLOSE_TAG_BEGIN, error))
	{
		SH_Text_free (html, NULL);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text, NULL);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, fragment->tag, error))
	{
		SH_Text_free (html, NULL);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text, NULL);
		}

		return NULL;
	}

	if (!SH_Text_append_string (html, CLOSE_TAG_END, error))
	{
		SH_Text_free (html, NULL);

		if (mode == WRAP)
		{
			SH_Text_free (indent_text, NULL);
		}

		return NULL;
	}

	if (mode == WRAP)
	{
		if (!SH_Text_append_string (html, NEWLINE, error))
		{
			SH_Text_free (html, NULL);
			SH_Text_free (indent_text, NULL);
			return NULL;
		}

		SH_Text_free (indent_text, NULL);
	}

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return html;
}
